"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the mining engine.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import socket
from hashlib import sha256

import urllib3.exceptions as urllib3_exceptions

import PyFunceble


class Mining:  # pylint: disable=too-many-instance-attributes
    """
    Manages the minig subsystem.
    """

    database = {}
    is_subject_present_cache = {}
    database_file = None

    authorized = False
    filename = None
    headers = {}

    def __init__(
        self, filename, mysql_db=None, parent_process=False
    ):  # pragma: no cover
        # We get the authorization to operate.
        self.authorized = self.authorization()
        self.database_file = ""
        # We save the file we are working with.
        self.filename = filename
        # Se create the current file namespace.
        self.database[self.filename] = {}
        # We share the state.
        self.parent = parent_process

        self.mysql_db = mysql_db

        self.table_name = self.get_table_name()

        PyFunceble.LOGGER.debug(f"Authorization: {self.authorized}")
        PyFunceble.LOGGER.debug(f"DB: {self.mysql_db}")
        PyFunceble.LOGGER.debug(f"Table Name: {self.table_name}")

        user_agent = PyFunceble.engine.UserAgent().get()

        if user_agent:
            # The user-agent is given.

            # We append the user agent to the header we are going to parse with
            # the request.
            self.headers = {"User-Agent": user_agent}

        if self.authorized:
            # We are authorized to operate.

            if PyFunceble.CONFIGURATION.db_type == "json":
                # We get the file we are going to save our data.
                self.database_file = (
                    PyFunceble.CONFIG_DIRECTORY
                    + PyFunceble.OUTPUTS.default_files.mining
                )

            PyFunceble.LOGGER.debug(f"DB (File): {self.database_file}")

            self.load()

    def __getitem__(self, index):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                if index in self.database[self.filename]:
                    return self.database[self.filename][index]

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "SELECT * "
                    "FROM {0} "
                    "WHERE file_path = %(file)s "
                    "AND subject = %(subject)s "
                ).format(self.table_name)

                with self.mysql_db.get_connection() as cursor:
                    cursor.execute(query, {"file": self.filename, "subject": index})

                    fetched = cursor.fetchall()

                    if fetched:
                        return [x["mined"] for x in fetched]

        return None

    def __setitem__(self, index, value):  # pylint: disable=too-many-branches
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                actual_value = self[index]

                if actual_value:
                    if isinstance(actual_value, dict):
                        if isinstance(value, dict):  # pragma: no cover
                            self.database[self.filename][index].update(value)
                        else:  # pragma: no cover
                            self.database[self.filename][index] = value
                    elif isinstance(actual_value, list):
                        if isinstance(value, list):
                            self.database[self.filename][index].extend(value)
                        else:  # pragma: no cover
                            self.database[self.filename][index].append(value)
                    else:  # pragma: no cover
                        self.database[self.filename][index] = value
                else:
                    if self.filename not in self.database:  # pragma: no cover
                        self.database[self.filename] = {}

                    self.database[self.filename][index] = value

                PyFunceble.LOGGER.info(
                    f"Inserted {repr(value)} into the subset of {repr(index)}"
                )
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "INSERT INTO {0} "
                    "(file_path, subject, mined, digest) "
                    "VALUES (%(file)s, %(subject)s, %(mined)s, %(digest)s)"
                ).format(self.table_name)

                with self.mysql_db.get_connection() as cursor:
                    for val in value:
                        digest = sha256(
                            bytes(self.filename + index + val, "utf-8")
                        ).hexdigest()

                        playload = {
                            "file": self.filename,
                            "subject": index,
                            "mined": val,
                            "digest": digest,
                        }
                        try:
                            cursor.execute(query, playload)

                            PyFunceble.LOGGER.info(
                                f"Inserted into the database: \n {playload}"
                            )
                        except self.mysql_db.errors:
                            pass

    def __delitem__(self, index):  # pragma: no cover
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                actual_value = self[index]

                if actual_value:
                    del self.database[self.filename][index]

                    PyFunceble.LOGGER.info(
                        "Cleaned the data related to "
                        f"{repr(index)} and {repr(self.filename)} "
                        f"from the database."
                    )
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "DELETE FROM {0} "
                    "WHERE file_path = %(file)s "
                    "AND subject = %(subject)s "
                ).format(self.table_name)

                with self.mysql_db.get_connection() as cursor:
                    cursor.execute(query, {"file": self.filename, "subject": index})

                    PyFunceble.LOGGER.info(
                        "Cleaned the data related to "
                        f"{repr(index)} and {repr(self.filename)} "
                        f"from the {repr(self.table_name)} table."
                    )

    @classmethod
    def authorization(cls):
        """
        Provides the operation authorization.
        """

        return PyFunceble.CONFIGURATION.mining

    @classmethod
    def get_history(cls, url):  # pragma: no cover
        """
        Gets the history of the given url.

        :param str url: An URL to call.

        :return: The list of links.
        :rtype: list
        """

        try:
            return PyFunceble.REQUESTS.get(
                url,
                headers=cls.headers,
                timeout=PyFunceble.CONFIGURATION.timeout,
                verify=PyFunceble.CONFIGURATION.verify_ssl_certificate,
                allow_redirects=True,
            ).history
        except (
            PyFunceble.REQUESTS.exceptions.ConnectionError,
            PyFunceble.REQUESTS.exceptions.Timeout,
            PyFunceble.REQUESTS.exceptions.InvalidURL,
            socket.timeout,
            urllib3_exceptions.InvalidHeader,
            UnicodeDecodeError,  # The probability that this happend in production is minimal.
        ):
            PyFunceble.LOGGER.exception()

            return []

    def get_table_name(self):
        """
        Returns the name of the table to use.
        """

        if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
            return self.mysql_db.tables["mining"]
        return "mining"

    def list_of_mined(self):
        """
        Provides the list of mined domains so that they can
        be tested.

        :return:
            The list of mined domains.

            The returned format is the following:

                ::

                    [
                        (index_to_delete_after_test, mined),
                        (index_to_delete_after_test, mined),
                        (index_to_delete_after_test, mined)
                    ]
        :rtype: list
        """

        PyFunceble.LOGGER.info(
            "Getting the list of previously mined data. (DATASET WONT BE LOGGED)"
        )

        # We initiate a variable which will return the result.
        result = []

        if self.authorized:
            # We are authorized to operate.

            if PyFunceble.CONFIGURATION.db_type == "json":

                for subject in self.database[self.filename].keys():
                    # We loop through the available list of status
                    # from the database.

                    for element in self[subject]:
                        # We then loop through the data associatied to
                        # the currently read status.

                        result.append((subject, element))
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = "SELECT * FROM {0} WHERE file_path = %(file)s".format(
                    self.table_name
                )

                with self.mysql_db.get_connection() as cursor:
                    cursor.execute(query, {"file": self.filename})

        # We return the result.
        return result

    def load(self):
        """
        Loads the content of the database file.
        """

        if self.authorized and PyFunceble.CONFIGURATION.db_type == "json":
            # We are authorized to operate.

            if PyFunceble.helpers.File(self.database_file).exists():
                # The database file exists.

                # We update the database with the content of the file.
                self.database.update(
                    PyFunceble.helpers.Dict().from_json_file(self.database_file)
                )

                PyFunceble.LOGGER.info(
                    "Database content loaded in memory. (DATASET WONT BE LOGGED)"
                )

    def save(self):
        """
        Saves the content of the database into the database file.
        """

        if (
            self.authorized
            and self.parent
            and PyFunceble.CONFIGURATION.db_type == "json"
        ):
            # We are authorized to operate.

            # We save the database into the file.
            PyFunceble.helpers.Dict(self.database).to_json_file(self.database_file)

            PyFunceble.LOGGER.info(f"Saved database into {repr(self.database_file)}.")

    # pylint: disable=too-many-branches
    def mine(self, subject, subject_type):  # pragma: no cover
        """
        Searches for domain or URL related to the original URL or domain.
        If some are found, we add them into the database.

        :param str subject: The subject we are working with.

        :param str subject_typ:
            The type of the subject.

            Can me one of the following:

                - :code:`url`

                - :code:`domain`
        """

        if self.authorized and not self[subject]:
            # We are authorized to operate.

            PyFunceble.LOGGER.info("Starting mining logic..")

            if subject_type == "domain":
                to_get = "http://{0}:80".format(subject)
            elif subject_type == "url":
                to_get = subject
            else:
                raise ValueError("Unknown subject type {0}".format(repr(subject_type)))

            history = self.get_history(to_get)

            PyFunceble.LOGGER.debug(f"(Not processed) Mined:\n{history}")

            for element in history:
                # We loop through the list of requests history.

                # We get the url from the currently read
                # request.
                url = element.url

                # We create a variable which will save the
                # local result.
                local_result = None

                if subject_type == "domain":
                    # We are working with domains.

                    # We validate and get the base of the URL we
                    # are working with.
                    local_result = PyFunceble.Check(url).is_url(return_base=True)
                elif subject_type == "url":
                    # We are working with URLs.

                    # We validate and get the full URL.
                    local_result = PyFunceble.Check(url).is_url(
                        return_base=False, return_formatted=True
                    )

                if local_result:
                    # The subject type is domain.

                    PyFunceble.LOGGER.debug(
                        f"{local_result} was successfully validated. "
                        "Removing possible Ports."
                    )

                    if subject_type == "domain":
                        # We are working with domain.

                        if local_result.endswith(":80"):
                            # The 80 port is present.

                            # We remove it.
                            local_result = local_result[: local_result.find(":80")]
                        elif local_result.endswith(":443"):
                            # The 443 port is present.

                            # We remove it.
                            local_result = local_result[: local_result.find(":443")]

                    if local_result != subject:
                        # The local result is differnt from the
                        # subject we are working with.

                        PyFunceble.LOGGER.debug(
                            f"Saving {repr(local_result)} into the subset of {repr(subject)}."
                        )

                        # We save into the database.
                        self[subject] = [local_result]

            # We save the database.
            self.save()

    def remove(self, subject, history_member):
        """
        Removes the given subject from the database assigned to the
        currently tested file.

        :param str subject: The subject we are working with.
        :param str history_member: The history member to delete.
        """

        if self.authorized:
            while True:
                actual_value = self[subject]

                if isinstance(actual_value, list) and history_member in actual_value:

                    if PyFunceble.CONFIGURATION.db_type == "json":
                        try:
                            actual_value.remove(history_member)

                            PyFunceble.LOGGER.info(
                                f"Removed {repr(history_member)} (mined) "
                                f"From the subset of {repr(subject)}."
                            )
                        except ValueError:  # pragma: no cover
                            pass
                    elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                        # We construct the query string.
                        query = (
                            "DELETE FROM {0} "
                            "WHERE file_path = %(file)s "
                            "AND subject = %(subject)s "
                            "AND mined = %(mined)s"
                        ).format(self.table_name)

                        with self.mysql_db.get_connection() as cursor:
                            cursor.execute(
                                query,
                                {
                                    "file": self.filename,
                                    "subject": subject,
                                    "mined": history_member,
                                },
                            )

                            PyFunceble.LOGGER.info(
                                "Cleaned the data related to "
                                f"{repr(subject)}, {repr(history_member)} (mined) and "
                                f"{repr(self.filename)} and from "
                                f"the {repr(self.table_name)} table."
                            )
                else:  # pragma: no cover
                    break

            if not self[subject]:  # pragma: no cover
                del self[subject]

            self.save()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the custom logs interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from datetime import datetime

import PyFunceble


class Logs:
    """
    Provide a clean and unique way to work with logs.
    Indeed, it's not good to have logs spread around the code :smile:

    :param str output: A path to the JSON file we are going to write.
    """

    def __init__(self, output=None):
        self.output = output
        self.current_time = str(datetime.now().timestamp())

    @classmethod
    def _get_content(cls, file):
        """
        Get and return the content of the given log file.

        :param str file: The file we have to get the content from.

        :return The content of the given file.
        :rtype: dict
        """

        if PyFunceble.helpers.File(file).exists():
            return PyFunceble.helpers.Dict().from_json_file(file)

        return {}

    @classmethod
    def _write_content(cls, content, file):
        """
        Write the content into the given file.

        :param str content: The dict to write.

        :param str file: The file to write.
        """

        if not PyFunceble.CONFIGURATION.no_files:
            if not isinstance(content, dict):
                content = {}

            PyFunceble.helpers.Dict(content).to_json_file(file)

    def whois(self, subject, record):
        """
        Logs the WHOIS record if needed.

        :param str subject: The currently tested subject.
        :param str record: The record to log.
        """

        if PyFunceble.CONFIGURATION.debug and PyFunceble.CONFIGURATION.logs:
            # The debug and the logs subsystem are activated.

            to_write = {self.current_time: {"domain": subject, "record": record}}

            if self.output:
                output = self.output
            else:
                output = PyFunceble.OUTPUT_DIRECTORY
                output += PyFunceble.OUTPUTS.parent_directory
                output += PyFunceble.OUTPUTS.logs.directories.parent
                output += PyFunceble.OUTPUTS.logs.filenames.whois

            current_content = self._get_content(output)
            current_content.update(to_write)

            PyFunceble.LOGGER.debug(f"WHOIS Record of {repr(subject)}:\n{to_write}")

            self._write_content(current_content, output)

    def expiration_date(self, subject, extracted):
        """
        Logs the extracted expiration date.

        :param str subject: The currently tested subject.
        :param str extracted: The extracted expiration date (from WHOIS record).
        """

        if PyFunceble.CONFIGURATION.logs:
            # The logs subsystem is activated.

            if PyFunceble.INTERN["referer"]:
                referer = PyFunceble.INTERN["referer"]
            else:
                referer = None

            to_write = {
                self.current_time: {
                    "domain": subject,
                    "expiration_date": extracted,
                    "whois_server": referer,
                }
            }

            if self.output:
                output = self.output
            else:
                output = PyFunceble.OUTPUT_DIRECTORY
                output += PyFunceble.OUTPUTS.parent_directory
                output += PyFunceble.OUTPUTS.logs.directories.parent
                output += PyFunceble.OUTPUTS.logs.filenames.date_format

            current_content = self._get_content(output)
            current_content.update(to_write)

            PyFunceble.LOGGER.critical(
                f"Wrong date format for {repr(subject)}:\n{to_write}"
            )

            self._write_content(current_content, output)

            if PyFunceble.CONFIGURATION.share_logs:
                # The logs sharing is activated.

                # And we share the logs with the api.
                PyFunceble.REQUESTS.post(
                    PyFunceble.LINKS.api_date_format,
                    data=to_write[self.current_time],
                    timeout=PyFunceble.CONFIGURATION.timeout,
                    verify=PyFunceble.CONFIGURATION.verify_ssl_certificate,
                    allow_redirects=False,
                )

    def referer_not_found(self, subject, extension):
        """
        Logs the case that the referer was not found.

        :param str subject: The currently tested subject.
        :param str extension: The extension of the domain we are testing.
        """

        if PyFunceble.CONFIGURATION.logs:
            # The logs subsystem is activated.

            to_write = {self.current_time: {"domain": subject, "extension": extension}}

            if self.output:
                output = self.output
            else:
                output = PyFunceble.OUTPUT_DIRECTORY
                output += PyFunceble.OUTPUTS.parent_directory
                output += PyFunceble.OUTPUTS.logs.directories.parent
                output += PyFunceble.OUTPUTS.logs.filenames.no_referer

            current_content = self._get_content(output)
            current_content.update(to_write)

            PyFunceble.LOGGER.critical(
                f"Referer not found for {repr(subject)}:\n{to_write}"
            )

            self._write_content(current_content, output)

            if PyFunceble.CONFIGURATION.share_logs:
                # The logs sharing is activated.

                # And we share the logs with the api.
                PyFunceble.REQUESTS.post(
                    PyFunceble.LINKS.api_no_referer,
                    data=to_write[self.current_time],
                    timeout=PyFunceble.CONFIGURATION.timeout,
                    verify=PyFunceble.CONFIGURATION.verify_ssl_certificate,
                    allow_redirects=False,
                )



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the auto-save engine.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from datetime import datetime, timedelta

import PyFunceble

from .ci import GitLabCI, TravisCI


class AutoSave:  # pragma: no cover  pylint: disable=too-few-public-methods
    """
    Provides the autosave interface.
    """

    # We set the varible which will save the global authorization to operate.
    authorized = None
    # We set the variable which will save the starting time.
    start_time = None
    # We set the variable which will save the end time.
    endtime = None
    # We set the variable which will process as a chache for the
    # time exceeed value.
    time_exceed = False

    def __init__(self, start_time=None):
        self.current_ci_engine = self.get_current_ci()
        PyFunceble.LOGGER.debug(f"Current CI engine: {self.current_ci_engine}")

        if self.current_ci_engine:
            self.authorized = True

            PyFunceble.LOGGER.debug(
                "CI (already) Initiated: %s" % "ci_initiated" not in PyFunceble.INTERN
            )
            if "ci_initiated" not in PyFunceble.INTERN:
                self.current_ci_engine.init()
                self.current_ci_engine.bypass()

                if start_time is None:
                    self.start_time = datetime.now()
                else:
                    self.start_time = datetime.fromtimestamp(int(start_time))

                self.end_time = self.start_time + timedelta(
                    minutes=int(PyFunceble.CONFIGURATION.ci_autosave_minutes)
                )

                PyFunceble.INTERN["auto_save"] = {
                    "start_time": self.start_time.timestamp(),
                    "end_time": self.end_time.timestamp(),
                }

                PyFunceble.LOGGER.debug(f"Start Time: {self.start_time}")
                PyFunceble.LOGGER.debug(f"End Time:  {self.end_time}")

                PyFunceble.INTERN["ci_initiated"] = True
            elif "auto_save" in PyFunceble.INTERN:
                self.start_time = datetime.fromtimestamp(
                    PyFunceble.INTERN["auto_save"]["start_time"]
                )
                self.end_time = datetime.fromtimestamp(
                    PyFunceble.INTERN["auto_save"]["end_time"]
                )

                PyFunceble.LOGGER.debug(f"Start Time (Shared): {self.start_time}")
                PyFunceble.LOGGER.debug(f"End Time (Shared):  {self.end_time}")

    @classmethod
    def get_current_ci(cls):
        """
        Provides the current CI to use.
        """

        for ci_engine in [TravisCI(), GitLabCI()]:
            if ci_engine.authorized:
                return ci_engine

        return None

    def is_time_exceed(self):
        """
        Checks if the end time is exceed.
        """

        if self.authorized and hasattr(self, "end_time"):
            # We are authorized to operate.

            if not self.time_exceed and datetime.now() >= self.end_time:
                # * We did not tested previously if the time exceed.
                # and
                # * The time exceed.

                # We update the time exceed marker.
                self.time_exceed = True
        else:
            self.time_exceed = False

            PyFunceble.LOGGER.debug(f"Time exceed: {self.time_exceed}")

        return self.time_exceed

    def process(self, test_completed=False):
        """
        Processes the autosave base on the current state of the test.

        :param bool test_completed: Tell us if we finished the test.
        """

        if self.authorized:
            # We are authorized to operate.

            if test_completed:
                # The test was completed.

                # We run the end commit.
                self.current_ci_engine.end_commit()
            elif self.is_time_exceed():
                # The current time excessed the minimal time for autosaving.

                # We run the not end commit.
                self.current_ci_engine.not_end_commit()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the IANA database generation tool.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""


from multiprocessing import Pool

import PyFunceble


class Iana:  # pragma: no cover pylint: disable=too-few-public-methods
    """
    Logic behind the update and usage of `iana-domains-db.json`
    """

    # We map the list of server which have to be set manually because
    # they are not present into the IANA Root Zone Database.
    manual_server = {
        "aaa": "whois.nic.aaa",
        "abb": "whois.nic.abb",
        "able": "whois.nic.able",
        "accenture": "whois.nic.accenture",
        "aetna": "whois.nic.aetna",
        "aig": "whois.nic.aig",
        "americanexpress": "whois.nic.americanexpress",
        "amex": "whois.nic.amex",
        "amica": "whois.nic.amica",
        "amsterdam": "whois.nic.amsterdam",
        "analytics": "whois.nic.analytics",
        "aramco": "whois.nic.aramco",
        "arte": "whois.nic.arte",
        "as": "whois.nic.as",
        "athleta": "whois.nic.athleta",
        "audible": "whois.nic.audible",
        "author": "whois.nic.author",
        "aws": "whois.nic.aws",
        "axa": "whois.nic.axa",
        "azure": "whois.nic.azure",
        "baby": "whois.nic.baby",
        "banamex": "whois.nic.banamex",
        "bananarepublic": "whois.nic.bananarepublic",
        "baseball": "whois.nic.baseball",
        "bharti": "whois.nic.bharti",
        "bing": "whois.nic.bing",
        "bloomberg": "whois.nic.bloomberg",
        "bm": "whois.afilias-srs.net",
        "book": "whois.nic.book",
        "booking": "whois.nic.booking",
        "bot": "whois.nic.bot",
        "buzz": "whois.nic.buzz",
        "bz": "whois.afilias-grs.net",
        "call": "whois.nic.call",
        "calvinklein": "whois.nic.calvinklein",
        "caravan": "whois.nic.caravan",
        "cartier": "whois.nic.cartier",
        "caseih": "whois.nic.caseih",
        "cbn": "whois.nic.cbn",
        "cbre": "whois.nic.cbre",
        "cd": "chois.nic.cd",
        "chase": "whois.nic.chase",
        "circle": "whois.nic.circle",
        "cisco": "whois.nic.cisco",
        "citadel": "whois.nic.citadel",
        "citi": "whois.nic.citi",
        "citic": "whois.nic.citic",
        "cm": "whois.netcom.cm",
        "coupon": "whois.nic.coupon",
        "crown": "whois.nic.crown",
        "crs": "whois.nic.crs",
        "deal": "whois.nic.deal",
        "dealer": "whois.nic.dealer",
        "dell": "whois.nic.dell",
        "dhl": "whois.nic.dhl",
        "discover": "whois.nic.discover",
        "dnp": "whois.nic.dnp",
        "doosan": "whois.nic.doosan",
        "duns": "whois.nic.duns",
        "dupont": "whois.nic.dupont",
        "earth": "whois.nic.earth",
        "energy": "whois.nic.energy",
        "epost": "whois.nic.epost",
        "everbank": "whois.nic.everbank",
        "farmers": "whois.nic.farmers",
        "fast": "whois.nic.fast",
        "ferrero": "whois.nic.ferrero",
        "fire": "whois.nic.fire",
        "fj": "whois.usp.ac.fj",
        "flickr": "whois.nic.flickr",
        "flir": "whois.nic.flir",
        "food": "whois.nic.food",
        "ford": "whois.nic.ford",
        "fox": "whois.nic.fox",
        "free": "whois.nic.free",
        "frontier": "whois.nic.frontier",
        "ftr": "whois.nic.ftr",
        "ga": "whois.my.ga",
        "gap": "whois.nic.gap",
        "gh": "whois.nic.gh",
        "gmo": "whois.nic.gmo",
        "got": "whois.nic.got",
        "grainger": "whois.nic.grainger",
        "grocery": "whois.nic.grocery",
        "guardian": "whois.nic.guardian",
        "gucci": "whois.nic.gucci",
        "hair": "whois.nic.hair",
        "hbo": "whois.nic.hbo",
        "health": "whois.nic.health",
        "homegoods": "whois.nic.homegoods",
        "homesense": "whois.nic.homesense",
        "honeywell": "whois.nic.honeywell",
        "hot": "whois.nic.hot",
        "hoteles": "whois.nic.hoteles",
        "hotels": "whois.nic.hotels",
        "hotmail": "whois.nic.hotmail",
        "hsbc": "whois.nic.hsbc",
        "htc": "whois.nic.htc",
        "hyatt": "whois.nic.hyatt",
        "ieee": "whois.nic.ieee",
        "iinet": "whois.nic.iinet",
        "imdb": "whois.nic.imdb",
        "int": "whois.iana.org",
        "intel": "whois.nic.intel",
        "intuit": "whois.nic.intuit",
        "ipiranga": " whois.nic.ipiranga",
        "ipirange": "whois.nic.ipiranga",
        "itau": "whois.nic.itau",
        "iwc": "whois.nic.iwc",
        "jetzt": "whois.nic.jetzt",
        "jlc": "whois.nic.jlc",
        "jmp": "whois.nic.jmp",
        "jnj": "whois.nic.jnj",
        "jot": "whois.nic.jot",
        "joy": "whois.nic.joy",
        "jpmorgan": "whois.nic.jpmorgan",
        "jprs": "whois.nic.jprs",
        "kinder": "whois.nic.kinder",
        "kindle": "whois.nic.kindle",
        "kpmg": "whois.nic.kpmg",
        "kpn": "whois.nic.kpn",
        "kred": "whois.nic.kred",
        "kw": "whois.nic.kw",
        "lanxess": "whois.nic.lanxess",
        "lc": "whois2.afilias-grs.net",
        "lifeinsurance": "whois.nic.lifeinsurance",
        "like": "whois.nic.like",
        "lilly": "whois.nic.lilly",
        "lincoln": "whois.nic.lincoln",
        "living": "whois.nic.living",
        "lk": "whois.nic.lk",
        "loft": "whois.nic.loft",
        "lupin": "whois.nic.lupin",
        "maif": "whois.nic.maif",
        "marshalls": "whois.nic.marshalls",
        "mattel": "whois.nic.mattel",
        "mcd": "whois.nic.mcd",
        "mcdonalds": "whois.nic.mcdonalds",
        "merckmsd": "whois.nic.merckmsd",
        "microsoft": "whois.nic.microsoft",
        "mint": "whois.nic.mint",
        "mlb": "whois.nic.mlb",
        "mobily": "whois.nic.mobily",
        "moi": "whois.nic.moi",
        "montblanc": "whois.nic.montblanc",
        "moto": "whois.nic.moto",
        "msd": "whois.nic.msd",
        "mtpc": "whois.nic.mtpc",
        "mutual": "whois.nic.mutual",
        "mutuelle": "whois.nic.mutuelle",
        "nagoya": "whois.nic.nagoya",
        "nba": "whois.nic.nba",
        "netflix": "whois.nic.netflix",
        "neustar": "whois.nic.neustar",
        "nfl": "whois.nic.nfl",
        "nhk": "whois.nic.nhk",
        "nike": "whois.nic.nike",
        "northwesternmutual": "whois.nic.northwesternmutual",
        "now": " whois.nic.now",
        "ntt": "whois.nic.ntt",
        "nyc": "whois.nic.nyc",
        "office": "whois.nic.office",
        "okinawa": "whois.nic.okinawa",
        "oldnavy": "whois.nic.oldnavy",
        "open": "whois.nic.open",
        "orientexpress": "whois.nic.orientexpress",
        "otsuka": "whois.nic.otsuka",
        "passagens": "whois.nic.passagens",
        "pay": "whois.nic.pay",
        "pfizer": "whois.nic.pfizer",
        "pharmacy": " whois.nic.pharmacy",
        "piaget": " whois.nic.piaget",
        "pictet": "whois.nic.pictet",
        "pin": "whois.nic.pin",
        "ping": "whois.nic.ping",
        "pramerica": "whois.nic.pramerica",
        "praxi": "whois.nic.praxi",
        "prime": "whois.nic.prime",
        "pru": "whois.nic.pru",
        "prudential": "whois.nic.prudential",
        "ps": "whois.pnina.ps",
        "qvc": "whois.nic.qvc",
        "read": "whois.nic.read",
        "realtor": "whois.nic.realtor",
        "ren": "whois.nic.ren",
        "rocher": "whois.nic.rocher",
        "room": "whois.nic.room",
        "rw": "whois.ricta.org.rw",
        "ryukyu": "whois.nic.ryukyu",
        "safe": "whois.nic.safe",
        "safety": "whois.nic.safety",
        "sakura": "whois.nic.sakura",
        "sapo": "whois.nic.sapo",
        "sas": "whois.nic.sas",
        "save": "whois.nic.save",
        "secure": "whois.nic.secure",
        "sener": "whois.nic.sener",
        "shaw": "whois.afilias-srs.net",
        "shop": "whois.nic.shop",
        "silk": "whois.nic.silk",
        "skype": "whois.nic.skype",
        "sl": "whois.nic.sl",
        "smile": "whois.nic.smile",
        "sohu": "whois.nic.sohu",
        "song": "whois.nic.song",
        "spot": "whois.nic.spot",
        "staples": "whois.nic.staples",
        "statefarm": "whois.nic.statefarm",
        "stream": "whois.nic.stream",
        "suzuki": "whois.nic.suzuki",
        "swiftcover": "whois.nic.swiftcover",
        "talk": "whois.nic.talk",
        "taobao": "whois.nic.taobao",
        "target": "whois.nic.target",
        "tjmaxx": "whois.nic.tjmaxx",
        "tjx": "whois.nic.tjx",
        "tkmaxx": "whois.nic.tkmaxx",
        "tmall": "whois.nic.tmall",
        "tokyo": "whois.nic.tokyo",
        "tube": "whois.nic.tube",
        "tunes": "whois.nic.tunes",
        "tushu": "whois.nic.tushu",
        "tvs": "whois.nic.tvs",
        "unicom": "whois.nic.unicom",
        "uno": "whois.nic.uno",
        "vivo": "whois.nic.vivo",
        "vuelos": "whois.nic.vuelos",
        "wanggou": "whois.nic.wanggou",
        "watches": "whois.nic.watches",
        "weather": "whois.nic.weather",
        "weatherchannel": "whois.nic.weatherchannel",
        "weir": "whois.nic.weir",
        "whois": "whois.nic.qpon",
        "windows": "whois.nic.windows",
        "winners": "whois.nic.winners",
        "wow": "whois.nic.wow",
        "xbox": "whois.nic.xbox",
        "xn--1ck2e1b": "whois.nic.xn--1ck2e1b",
        "xn--2scrj9c": "whois.inregistry.net",
        "xn--3hcrj9c": "whois.inregistry.net",
        "xn--45br5cyl": "whois.inregistry.net",
        "xn--45brj9c": "whois.inregistry.net",
        "xn--8y0a063a": "whois.nic.xn--8y0a063a",
        "xn--bck1b9a5dre4c": "whois.nic.xn--bck1b9a5dre4c",
        "xn--cck2b3b": "whois.nic.xn--cck2b3b",
        "xn--czr694b": "whois.nic.xn--czr694b",
        "xn--e1a4c": "whois.eu",
        "xn--eckvdtc9d": "whois.nic.xn--eckvdtc9d",
        "xn--fct429k": "whois.nic.xn--fct429k",
        "xn--fpcrj9c3d": "whois.inregistry.net",
        "xn--fzc2c9e2c": "whois.nic.lk",
        "xn--g2xx48c": "whois.nic.xn--g2xx48c",
        "xn--gckr3f0f": "whois.nic.xn--gckr3f0f",
        "xn--gecrj9c": "whois.inregistry.net",
        "xn--gk3at1e": "whois.nic.xn--gk3at1e",
        "xn--h2breg3eve": "whois.inregistry.net",
        "xn--h2brj9c": "whois.inregistry.net",
        "xn--h2brj9c8c": "whois.inregistry.net",
        "xn--imr513n": "whois.nic.xn--imr513n",
        "xn--jvr189m": "whois.nic.xn--jvr189m",
        "xn--kpu716f": "whois.nic.xn--kpu716f",
        "xn--mgba3a3ejt": "whois.nic.xn--mgba3a3ejt",
        "xn--mgbb9fbpob": "whois.nic.xn--mgbb9fbpob",
        "xn--mgbbh1a": "whois.inregistry.net",
        "xn--mgbbh1a71e": "whois.inregistry.net",
        "xn--mgbgu82a": "whois.inregistry.net",
        "xn--nyqy26a": "whois.nic.xn--nyqy26a",
        "xn--otu796d": "whois.nic.xn--otu796d",
        "xn--pbt977c": "whois.nic.xn--pbt977c",
        "xn--rhqv96g": "whois.nic.xn--rhqv96g",
        "xn--rovu88b": "whois.nic.xn--rovu88b",
        "xn--rvc1e0am3e": "whois.inregistry.net",
        "xn--s9brj9c": "whois.inregistry.net",
        "xn--ses554g": "whois.registry.knet.cn",
        "xn--wgbh1c": "whois.dotmasr.eg",
        "xn--xkc2al3hye2a": "whois.nic.lk",
        "xn--xkc2dl3a5ee0h": "whois.inregistry.net",
        "yahoo": "whois.nic.yahoo",
        "yamaxun": "whois.nic.yamaxun",
        "yandex": "whois.nic.yandex",
        "yokohama": "whois.nic.yokohama",
        "you": "whois.nic.you",
        "za": "whois.registry.net.za",
        "zappos": "whois.nic.zappos",
        "zero": "whois.nic.zero",
        "zippo": "whois.nic.zippo",
    }

    def __init__(self):
        # We get the destination of the constructed IANA database.
        self.destination = (
            PyFunceble.CONFIG_DIRECTORY + PyFunceble.OUTPUTS.default_files.iana
        )

        if PyFunceble.helpers.File(self.destination).exists():
            # The destination exist.

            # We get its content.
            self.database = PyFunceble.helpers.Dict().from_json_file(self.destination)
        else:
            # The destination does not exist.

            # We initiate the local variable which will save the content of the database.
            self.database = {}

        # We initiate the URL to the IANA Root Zone Database page.
        self.iana_url = "https://www.iana.org/domains/root/db"

    def load(self):
        """
        Initiate the IANA database if it is not the case.
        """

        if "database" not in PyFunceble.INTERN or not PyFunceble.INTERN["database"]:
            # The global database is empty, None or does not exist.

            # We update it with the database content.
            PyFunceble.INTERN.update({"database": self.database.copy()})

    @classmethod
    def _get_referer(cls, extension):
        """
        Return the referer for the given extension.

        :param str extension: A valid domain extension.

        :return: The whois server to use to get the WHOIS record.
        :rtype: str
        """

        PyFunceble.LOGGER.info(f"Trying to find the referer of {repr(extension)}.")

        # We get the  whois record related to the domain extension we are currently
        # working with.
        iana_record = PyFunceble.lookup.Whois(
            "hello.{}".format(extension), PyFunceble.CONFIGURATION.iana_whois_server
        ).request()

        if iana_record and "refer" in iana_record:
            # The record is not empty.

            PyFunceble.LOGGER.info("Referer probably present. Trying to extract it.")

            # We initiate a regex which will extract the referer.
            regex_referer = r"(?s)refer\:\s+([a-zA-Z0-9._-]+)\n"

            # We try to extract the referer.
            matched = PyFunceble.helpers.Regex(regex_referer).match(
                iana_record, return_match=True, group=1
            )

            PyFunceble.LOGGER.debug(f"Extracted: {repr(matched)}")

            if matched:
                # The referer was extracted successfully.

                # We return the matched referer.
                return matched

        PyFunceble.LOGGER.info(
            "Could not extract nor find the referer from record. "
            "Trying to find it into our local entries."
        )

        # * The referer was not extracted successfully.
        # or
        # * The iana record is empty.

        if extension in cls.manual_server:
            # The extension is in the list of manual entries.

            PyFunceble.LOGGER.debug(
                f"Found local entry for {repr(extension)}: "
                f"{repr(cls.manual_server[extension])}"
            )

            # We return the server which we set manually.
            return cls.manual_server[extension]

        # We return None because we weren't able to get the server to call for
        # the given extension.
        return None

    @classmethod
    def _check_referer(cls, extension, referer):
        """
        Check if the given referer is reachable

        :param str extension: The extension the referer refers to.
        :param str referer: The referer to check.

        :rtype: bool
        """

        return (
            PyFunceble.lookup.Whois(
                "hello.{0}".format(extension), server=referer
            ).request()
            is not None
        )

    @classmethod
    def _get_extension_and_referer_from_block(cls, block):
        """
        Extract the extention from the given HTML block.
        Plus get its referer.

        :param str block: An HTML block.
        """

        # We extract the different extension from the currently readed line.
        regex_valid_extension = r"(/domains/root/db/)(.*)(\.html)"

        if "/domains/root/db/" in block:
            # The link is in the line.

            # We try to extract the extension.
            matched = PyFunceble.helpers.Regex(regex_valid_extension).match(
                block, return_match=True, rematch=True
            )[1]

            if matched:
                # The extraction is not empty or None.

                # We get the referer.
                referer = cls._get_referer(matched)

                if not referer:
                    referer = "whois.nic.{0}".format(matched)

                if cls._check_referer(matched, referer):
                    return matched, referer, True

                return matched, referer, False

        return None, None, None

    def update(self):
        """
        Update the content of `iana-domains-db.json` file.
        """

        if not PyFunceble.CONFIGURATION.quiet:
            # The quiet mode is not activated.

            # We print on screen what we are doing.
            print(
                "Update of {0}".format(PyFunceble.OUTPUTS.default_files.iana), end=" "
            )

        upstream_lines = (
            PyFunceble.helpers.Download(self.iana_url)
            .text()
            .split('<span class="domain tld">')
        )

        with Pool(PyFunceble.CONFIGURATION.maximal_processes) as pool:
            already_checked = []
            for extension, referer, referer_checked in pool.map(
                self._get_extension_and_referer_from_block, upstream_lines
            ):
                if (extension is not None and referer) and (
                    referer_checked or referer in already_checked
                ):
                    if (
                        extension not in self.database
                        or self.database[extension] != referer
                    ):
                        # We add the extension to the databae.
                        self.database[extension] = referer

                already_checked.append(referer)

        # We save the content of the constructed database.
        PyFunceble.helpers.Dict(self.database).to_json_file(self.destination)

        if not PyFunceble.CONFIGURATION.quiet:
            # The quiet mode is not activated.

            # We indicate that the work is done without any issue.
            print(PyFunceble.INTERN["done"] + "\n")



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the DNS lookup interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from socket import IPPROTO_TCP, gaierror, getaddrinfo, gethostbyaddr, herror

import dns.resolver
import dns.reversename
from dns.exception import DNSException

import PyFunceble


class DNSLookup:  # pylint: disable=too-few-public-methods
    """
    DNS lookup interface.

    :param str subject: The subject we are working with.
    :param dns_server: The DNS server we are working with.
    :type dns_server: list|tuple|str
    :param int lifetime: Set the lifetime of a query.
    """

    def __init__(self, dns_server=None, lifetime=3, tcp=False):
        self.given_dns_server = dns_server
        self.resolver = self.__get_resolver(dns_server)

        self.update_nameserver(dns_server)
        self.update_lifetime(lifetime)
        self.tcp = tcp

        PyFunceble.LOGGER.debug(
            f"DNS Resolver Nameservers: {self.resolver.nameservers}"
        )
        PyFunceble.LOGGER.debug(
            f"DNS Resolver (Nameservers) Port: {self.resolver.nameserver_ports}"
        )
        PyFunceble.LOGGER.debug(f"DNS Resolver Port: {self.resolver.port}")
        PyFunceble.LOGGER.debug(f"DNS Resolver timeout: {self.resolver.timeout}")
        PyFunceble.LOGGER.debug(f"DNS Resolver lifetime: {self.resolver.lifetime}")
        PyFunceble.LOGGER.debug(f"DNS Resolver over TCP: {self.tcp}")

        PyFunceble.INTERN["dns_lookup"] = {
            "resolver": self.resolver,
            "given_dns_server": dns_server,
        }

    @classmethod
    def __get_resolver(cls, dns_server):
        """
        Provides the configured dns resolver.
        """

        if dns_server:
            # A dns server is given.

            PyFunceble.LOGGER.info(
                f"DNS Server explicitly given, generating a new configurable resolver."
            )
            PyFunceble.LOGGER.debug(f"Given DNS server: {dns_server}")

            # We initiate and configure the resolver.
            resolver = dns.resolver.Resolver(configure=False)

            return resolver

        # A dns server is not given.

        PyFunceble.LOGGER.info(
            f"DNS Server not explicitly given, providing the systemwide one."
        )

        # We configure everything with what the OS gives us.
        return dns.resolver.Resolver()

    def update_lifetime(self, lifetime):
        """
        Updates the lifetime of a query.
        """

        if isinstance(lifetime, (int, float)):
            if lifetime == self.resolver.timeout:
                self.resolver.lifetime = float(lifetime) + 2.0
            else:
                self.resolver.lifetime = float(lifetime)
        else:
            self.resolver.lifetime = self.resolver.timeout + 2.0

    def update_nameserver(self, nameserver):
        """
        Updates the nameserver to query.
        """

        if nameserver:
            nameserver = (
                [nameserver]
                if not isinstance(nameserver, (list, tuple))
                else nameserver
            )

            self.resolver.nameserver_ports = {
                z: y
                for x, y in self.__get_server_and_port_from(nameserver)
                for z in self.__get_dns_servers_from(x)
            }

            self.resolver.nameservers = list(self.resolver.nameserver_ports.keys())

            PyFunceble.LOGGER.info(
                f"Switched Resolver nameserver to: {self.resolver.nameservers}"
            )
            PyFunceble.LOGGER.info(
                f"Switched Resolver nameserver port to: {self.resolver.nameserver_ports}"
            )

    def __get_server_and_port_from(self, inputed_dns):  # pragma: no cover
        """
        Given a list or an input representing dns server,
        we split the server from the port.

        :param inputed_dns: The inputed DNS server(s).
        :type input: str, list, tuple

        :return:
            A tuple with:

            (The DNS, the port)
        :rtype: tuple
        """

        if isinstance(inputed_dns, (list, tuple)):
            result = []

            for dns_server in inputed_dns:
                result.append(self.__get_server_and_port_from(dns_server))

            return result

        if ":" in inputed_dns:
            splited = inputed_dns.split(":")

            if len(splited) > 1 and splited[-1] and splited[-1].isdigit():
                return ":".join(splited[:-1]), int(splited[-1])

            if not splited[-1]:
                return ":".join(splited[:-1]), 53

        return inputed_dns, 53

    def __get_dns_servers_from(self, inputed_dns):  # pragma: no cover
        """
        Given a list or an input representing dns servers,
        we ensure that we have a list of a string which
        represent an IP.

        :param input: The inputed DNS server(s).
        :type input: str, list, tuple

        :return: The list of dns server to use.
        :rtype: list
        """

        if isinstance(inputed_dns, (list, tuple)):
            result = []
            for dns_server in inputed_dns:
                result.extend(self.__get_dns_servers_from(dns_server))

            return result

        result = []

        try:
            if PyFunceble.Check(inputed_dns).is_domain():
                result.extend(
                    [x.address for x in dns.resolver.Resolver().query(inputed_dns)]
                )
            else:
                result.append(inputed_dns)
        except DNSException:
            result.extend(dns.resolver.get_default_resolver().nameservers)

        return result

    def a_record(self, subject, tcp=None):
        """
        Return the A record of the given subject (if found).

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :return: A list of A record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting A record of {repr(subject)}")
            # We get the A record of the given subject.
            result = [str(x) for x in self.resolver.query(subject, "A", tcp=tcp)]
            PyFunceble.LOGGER.info(f"Could get A record of {repr(subject)}: {result}")

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get A record of {repr(subject)}")

        return None

    def aaaa_record(self, subject, tcp=None):
        """
        Return the AAAA record of the given subject (if found).

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :return: A list of A record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting AAAA record of {repr(subject)}")
            # We get the A record of the given subject.
            result = [str(x) for x in self.resolver.query(subject, "AAAA", tcp=tcp)]
            PyFunceble.LOGGER.info(
                f"Could get AAAA record of {repr(subject)}: {result}"
            )

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get AAAA record of {repr(subject)}")

        return None

    def cname_record(self, subject, tcp=None):
        """
        Return the CNAME record of the given subject (if found).

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :return: A list of CNAME record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting CNAME record of {repr(subject)}")
            # We get the A record of the given subject.
            result = [str(x) for x in self.resolver.query(subject, "CNAME", tcp=tcp)]
            PyFunceble.LOGGER.info(
                f"Could get CNAME record of {repr(subject)}: {result}"
            )

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get CNAME record of {repr(subject)}")

        return None

    def dname_record(self, subject, tcp=None):  # pragma: no cover
        """
        Return the DNAME record of the given subject (if found).

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :return: A list of DNAME record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting DNAME record of {repr(subject)}")
            # We get the A record of the given subject.
            result = [str(x) for x in self.resolver.query(subject, "DNAME", tcp=tcp)]
            PyFunceble.LOGGER.info(
                f"Could get DNAME record of {repr(subject)}: {result}"
            )

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get DNAME record of {repr(subject)}")

        return None

    def mx_record(self, subject, tcp=None):
        """
        Return the MX record of the given subject (if found).

        :param str subject: The subject we are working with.

        :return: A list of MX record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting MX record of {repr(subject)}")
            # We get the MX record of the given subject.
            result = [str(x) for x in self.resolver.query(subject, "MX", tcp=tcp)]

            PyFunceble.LOGGER.info(f"Could get MX record of {repr(subject)}: {result}")

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get MX record of {repr(subject)}")

        return None

    def ns_record(self, subject, tcp=None):
        """
        Return the NS record of the given subject (if found).

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :return: A list of NS record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting NS record of {repr(subject)}")
            # We get the NS record of the given subject.
            result = [str(x) for x in self.resolver.query(subject, "NS", tcp=tcp)]
            PyFunceble.LOGGER.info(f"Could get NS record of {repr(subject)}: {result}")

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get NS record of {repr(subject)}")

        return None

    def txt_record(self, subject, tcp=None):
        """
        Return the TXT record of the given subject (if found).

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :return: A list of TXT record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting TXT record of {repr(subject)}")
            # We get the TXT record of the given subject.
            result = [str(x) for x in self.resolver.query(subject, "TXT", tcp=tcp)]
            PyFunceble.LOGGER.info(f"Could get TXT record of {repr(subject)}: {result}")

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get TXT record of {repr(subject)}")

        return None

    def ptr_record(self, subject, reverse_name=True, tcp=None):
        """
        Return the PTR record of the given subject (if found).

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :return: A list of PTR record(s).
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"Query over TCP: {self.tcp}")

        try:
            PyFunceble.LOGGER.info(f"Getting TXT record of {repr(subject)}")
            if reverse_name:
                # We get the reverse name we are going to request.
                to_request = dns.reversename.from_address(subject)
            else:
                to_request = subject

            # We get the PTR record of the currently read A record.
            result = [str(x) for x in self.resolver.query(to_request, "PTR", tcp=tcp)]
            PyFunceble.LOGGER.info(f"Could get PTR record of {repr(subject)}: {result}")

            return result
        except DNSException:
            PyFunceble.LOGGER.error(f"Could not get PTR record of {repr(subject)}")

        return None

    @classmethod
    def get_addr_info(cls, subject):  # pragma: no cover
        """
        Get and return the information of the given subject (address).

        :param str subject: The subject we are working with.

        :return: A list of address.
        :rtype: list, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        try:  # pragma: no cover
            # We request the address information.
            req = getaddrinfo(subject, 80, proto=IPPROTO_TCP)

            # We format the addr infos.
            return [x[-1][0] for x in req]
        except (gaierror, OSError, herror, UnicodeError):
            pass

        return None

    @classmethod
    def get_host_by_addr(cls, subject):  # pragma: no cover
        """
        Get and return the host of the given subject (address).

        :param str subject: The subject we are working with.

        :return:
            A dict in the following format or :code:`None`.

            ::

                {
                    "hostname": "",
                    "aliases": [],
                    "ips": []
                }

        :rtype: dict, None
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        try:
            # We get the host by addr.
            req = gethostbyaddr(subject)

            # And we format the result.
            return {"hostname": req[0], "aliases": req[1], "ips": req[2]}

        except (gaierror, OSError, herror):
            pass

        return None

    @classmethod
    def is_record_present_in_result(cls, to_check, result, allow_empty=False):
        """
        Checks if the given record type is in the result.

        :param to_check:
            The record to check the presence.
        :type to_check: list, str, tuple
        :param dict result:
            The result to work with.
        :param bool allow_empty:
            Tell to if we allow and empty result as present.

        :rtype: bool
        """

        if isinstance(to_check, (list, tuple)):
            return any(
                [
                    cls.is_record_present_in_result(x, result, allow_empty=allow_empty)
                    for x in to_check
                ]
            )

        if to_check in result:
            if not allow_empty and not result[to_check]:
                PyFunceble.LOGGER.debug(
                    f"{to_check} record is not in result:\n{result}"
                )
                return False
            PyFunceble.LOGGER.debug(f"{to_check} record is in result:\n{result}")
            return True
        PyFunceble.LOGGER.debug(f"{to_check} record is not in result:\n{result}")
        return False

    def __request_complete_not_ip(self, subject, tcp=None):  # pragma: no cover
        """
        Requests and provides the complete DNS spectrum.

        :rtype: dict
        """

        result = {}

        # We get the A record of the given subject.
        result["A"] = self.a_record(subject, tcp=tcp)

        # We get the AAAA record of the given subject.
        result["AAAA"] = self.aaaa_record(subject, tcp=tcp)

        # We get the CNAME record of the given subject.
        result["CNAME"] = self.cname_record(subject, tcp=tcp)

        # We get the DNAME record of the given subject.
        result["DNAME"] = self.dname_record(subject, tcp=tcp)

        # We get the MX record of the given subject.
        result["MX"] = self.mx_record(subject, tcp=tcp)

        # We get the TXT record of the given subject.
        result["TXT"] = self.txt_record(subject, tcp=tcp)

        if self.is_record_present_in_result("A", result):
            # We could get some A record(s).

            # We initiate the PTR.
            result["PTR"] = []

            for a_result in result["A"]:
                # We loop through the list of A records.

                if "." not in a_result:  # pragma: no cover
                    # There is no "." in the currently
                    # read A record.

                    # We continue the loop.
                    continue

                try:
                    # We get the PTR record of the currently read A record.
                    result["PTR"].extend(self.ptr_record(a_result, tcp=tcp))
                except TypeError:  # pragma: no cover
                    pass

            if not all(result["PTR"]):  # pragma: no cover
                # No PTR record was found.

                # We delete the PTR entry.
                del result["PTR"]

        return result

    def __request_not_ip(self, subject, complete=False, tcp=None):  # pragma: no cover
        """
        Handle the request for a subject which is not an IP.

        :rtype: dict
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.debug(f"{repr(subject)} is not IP. Requesting record.")

        result = {}

        # We get the NS record of the given subject.
        result["NS"] = self.ns_record(subject, tcp=tcp)

        if complete:  # pragma: no cover
            result = PyFunceble.helpers.Merge(
                self.__request_complete_not_ip(subject, tcp=tcp)
            ).into(result)

        if not self.is_record_present_in_result(["CNAME", "NS"], result):
            # As sometime we may not have a NS but a CNAME, we handle that case.
            result["CNAME"] = self.cname_record(subject, tcp=tcp)

        if not self.is_record_present_in_result(["CNAME", "DNAME", "NS"], result):
            # Same but with DNAME.
            result["DNAME"] = self.dname_record(subject, tcp=tcp)

        if not self.is_record_present_in_result(["A", "CNAME", "DNAME", "NS"], result):
            # Same but with A.
            result["A"] = self.a_record(subject, tcp=tcp)

        if not self.is_record_present_in_result(
            ["A", "AAAA", "CNAME", "DNAME", "NS"], result
        ):
            # Same but with A.
            result["AAAA"] = self.aaaa_record(subject, tcp=tcp)

        # We get the list of index to delete.
        to_delete = [x for x in result if not result[x]]

        for index in to_delete:
            # We loop through the list of index to delete.

            # And we delete them.
            del result[index]

        if not result and complete:  # pragma: no cover
            # We could not get DNS records about the given subject.

            # We get the addr infos.
            result["addr_info"] = self.get_addr_info(subject)

            if not result["addr_info"]:
                # The addr_info index is empty.

                # We delete it.
                del result["addr_info"]
        elif result:
            result["nameservers"] = self.resolver.nameservers

        PyFunceble.LOGGER.debug(
            f"Records for {repr(subject)} (with {self.resolver.nameservers}):\n{result}"
        )

        return result

    def __request_ip(self, subject, tcp=None):  # pragma: no cover
        """
        Handle the request for a subject which is an IP.

        :param str subject: The subject we are working with.
        :param bool tcp: Tell us to use TCP for query.

        :rtype: dict
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        PyFunceble.LOGGER.info(f"{repr(subject)} is an IP. Requesting PTR record.")

        result = {}

        # We get the PTR record of the given subject.
        result["PTR"] = self.ptr_record(subject, tcp=tcp)

        if "PTR" not in result or not result["PTR"]:
            del result["PTR"]

            PyFunceble.LOGGER.error(f"PTR record for {repr(subject)} not found.")
        elif result["PTR"]:
            result["nameservers"] = self.resolver.nameservers

        if not result:
            # We could not get DNS records about the given subject.

            PyFunceble.LOGGER.info(f"Getting hosts by addr for {repr(subject)}")

            result = self.get_host_by_addr(subject)

        PyFunceble.LOGGER.debug(f"Request result: \n{result}")

        return result

    def request(self, subject, complete=False, tcp=None):  # pragma: no cover
        """
        Perform a DNS lookup/requests.

        :param str subject: The subject we are working with.
        :param bool complete: Tell us to return as many result as possible.
        :param bool tcp: Tell us to use TCP for query.

        :return:
            A dict with following index if the given subject is not registered into the
            given DNS server. (More likely local subjects).

                ::

                    {
                        "hostname": "",
                        "aliases": [],
                        "ips": []
                    }

            A dict with following index for everything else (and if found).

                ::

                    {
                        "nameservers: [],
                        "A": [],
                        "AAAA": [],
                        "CNAME": [],
                        "DNAME": [],
                        "MX": [],
                        "NS": [],
                        "TXT": [],
                        "PTR": []
                    }

        :rtype: dict
        :raise ValueError: When a non string :code:`subject` is given.
        """

        if not subject or not isinstance(subject, str):
            raise ValueError(f"<subject> must be of type {str} and not empty.")

        if tcp is None:
            tcp = self.tcp

        result = {}

        if not PyFunceble.Check(subject).is_ip():
            # We are looking for something which is not an IPv4 or IPv6.

            temp_result = self.__request_not_ip(subject, complete=complete, tcp=tcp)

            if isinstance(temp_result, dict):
                result.update(temp_result)
        else:
            # We are working with something which is an IPv4 or IPv6.

            temp_result = self.__request_ip(subject, tcp=tcp)

            if isinstance(temp_result, dict):
                result.update(temp_result)

        return result



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the printing interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from collections import OrderedDict
from datetime import datetime

from colorama import Back, Fore, Style

import PyFunceble


class Prints:
    """
    Print data on screen and into a file if needed.
    Template Possibilities: Percentage, Less, HTTP and any status you want.

    :param list to_print: The list of data to print.

    :param str template:
        The template to use.

        .. note::
            Available templates:

                - :code:`Percentage`
                - :code:`Less`
                - :code:`HTTP`
                - any of the official status.

    :param str output_file: The path to the file to write.

    :param bool only_on_file:
        Tell us if we only have to print on file and not on screen.
    """

    def __init__(self, to_print, template, output_file=None, only_on_file=False):
        # We get the template.
        self.template = template
        # We get the output.
        self.output = output_file
        # We get the data to print.
        self.data_to_print = to_print
        # We get the state of the only on file.
        self.only_on_file = only_on_file

        # We initiate the variable which will save the list of header.
        # Note: We initiate an Ordered Dict because we want to keep
        # the order.
        self.headers = OrderedDict()

        PyFunceble.LOGGER.debug(f"Template: {self.template}")
        PyFunceble.LOGGER.debug(f"Destination: {self.output}")
        PyFunceble.LOGGER.debug(f"Data to print:\n{self.data_to_print}")
        PyFunceble.LOGGER.debug(f"Only print on file: {self.only_on_file}")

        # We iniate the Generic header and the spacement of each colomns.
        self.headers["Generic"] = OrderedDict(
            zip(
                [
                    "Subject",
                    "Status",
                    "Expiration Date",
                    "Source",
                    "HTTP Code",
                    "Analyze Date",
                ],
                [100, 11, 17, 10, 10, 20],
            )
        )

        # We iniate the official UP header and the spacement of each colomns.
        self.headers[PyFunceble.STATUS.official.up] = OrderedDict(
            zip(
                ["Subject", "Expiration Date", "Source", "HTTP Code", "Analyze Date"],
                [100, 17, 10, 10, 20],
            )
        )

        # We iniate the official VALID header and the spacement of each colomns.
        self.headers[PyFunceble.STATUS.official.valid] = OrderedDict(
            zip(["Subject", "Source", "Analyze Date"], [100, 10, 20])
        )

        # We iniate the official SANE header and the spacement of each colomns.
        self.headers[PyFunceble.STATUS.official.sane] = OrderedDict(
            zip(["Subject", "Source", "Analyze Date"], [100, 10, 20])
        )

        # We iniate the official MALICIOUS header and the spacement of each colomns.
        self.headers[PyFunceble.STATUS.official.malicious] = OrderedDict(
            zip(["Subject", "Source", "Analyze Date"], [100, 10, 20])
        )

        # We iniate the official DOWN header and the spacement of each colomns.
        self.headers[PyFunceble.STATUS.official.down] = OrderedDict(
            zip(
                [
                    "Subject",
                    "WHOIS Server",
                    "Status",
                    "Source",
                    "HTTP Code",
                    "Analyze Date",
                ],
                [100, 35, 11, 10, 10, 20],
            )
        )

        # We iniate the official INVALID header and the spacement of each colomns.
        self.headers[PyFunceble.STATUS.official.invalid] = OrderedDict(
            zip(["Subject", "Source", "HTTP Code", "Analyze Date"], [100, 10, 10, 20])
        )

        # We iniate the official LESS header and the spacement of each colomns.
        self.headers["Less"] = OrderedDict(
            zip(["Subject", "Status", "HTTP Code"], [100, 11, 10])
        )

        try:
            # We iniate the official Simple header and the spacement of each colomns.
            self.headers["Simple"] = OrderedDict(
                zip(["Subject", "Status"], [len(self.data_to_print[0]), 11])
            )
        except TypeError:
            pass

        # We iniate the official Percentage header and the spacement of each colomns.
        self.headers["Percentage"] = OrderedDict(
            zip(["Status", "Percentage", "Numbers"], [11, 12, 12])
        )

        # We iniate the official HTTP header and the spacement of each colomns.
        self.headers["HTTP"] = OrderedDict(
            zip(["Subject", "Status", "HTTP Code", "Analyze Date"], [100, 11, 10, 20])
        )

        # We initiate a variable which will save the currently in use header.
        self.currently_used_header = {}

        # We initate a instance of the file output.
        self.file_output_instance = PyFunceble.helpers.File(self.output)

    def before_header(self):
        """
        Print informations about PyFunceble and the date of generation of a file
        into a given path, if doesn't exist.
        """

        if (
            not PyFunceble.CONFIGURATION.no_files
            and self.output
            and not self.file_output_instance.exists()
        ):

            # * We are allowed to generate files.
            # and
            # * And output is given.
            # and
            # * The given output does not exist.

            # We initiate the information about what generated the file.
            link = "# Generated by {0} (v{1}) / {2}\n".format(
                PyFunceble.NAME,
                PyFunceble.VERSION.split()[0],
                PyFunceble.abstracts.Infrastructure.REPO_LINK,
            )

            # We initiate the information about the generation date of this file.
            date_of_generation = (
                "# Date of generation: %s\n\n" % datetime.now().isoformat()
            )

            # We initiate a variable which will save the list of
            # templates which have to meet in order to write the before
            # header informations.
            authorized_templates = [
                "Generic_File",
                PyFunceble.STATUS.official.up,
                PyFunceble.STATUS.official.down,
                PyFunceble.STATUS.official.invalid,
                PyFunceble.STATUS.official.valid,
                PyFunceble.STATUS.official.sane,
                PyFunceble.STATUS.official.malicious,
                "Less",
            ]

            if self.template in authorized_templates:
                # The current header is in our list of authorized templated.

                # We get the header.
                header = (
                    self.header_constructor(self.currently_used_header, None)[0] + "\n"
                )

                PyFunceble.LOGGER.debug(f"HEADER: {header}")

            try:
                # We try to print the link, the date of generation and the header in the
                # given file.
                self.file_output_instance.write(link + date_of_generation + header)
            except UnboundLocalError:
                # We don't have any header.

                # We print the link and the date in the given file.
                self.file_output_instance.write(link + date_of_generation)

            PyFunceble.LOGGER.info(
                f"Created the {self.file_output_instance.path} file with the header."
            )

    @classmethod
    def header_constructor(
        cls, data_to_print, header_separator="-", column_separator=" "
    ):
        """
        Construct header of the table according to template.

        :param list data_to_print:
            The list of data to print into the header of the table.

        :param str header_separator:
            The separator to use between the table header and our data.

        :param str colomn_separator: The separator to use between each colomns.

        :return: The data to print in a list format.
        :rtype: list
        """

        # We initiate a variable which will save the header data.
        header_data = []

        # We initiate a variable which will save the header sizes.
        header_size = ""

        # We initiate the glue to set before the size.
        before_size = "%-"

        # We initiate the glue to set after the size.
        after_size = "s"

        if header_separator:
            # The header separator is not empty.

            # We initiate a variable which will save the list of
            # separator data.
            header_separator_data = []

        # We get the length of the data to print.
        length_data_to_print = len(data_to_print) - 1

        # We initiate an iterator.
        i = 0

        for data in data_to_print:
            # We loop through the list of data.

            # We get the size of the currently read data.
            size = data_to_print[data]

            # We append the data to the header data list.
            header_data.append(data)

            # We construct the header size.
            # Note: our header size is formatted line %s-sizes
            # (the s at the end is part of the formatting.)
            header_size += before_size + str(size) + after_size

            if i < length_data_to_print:
                # The iterator is less than the length of data to print.

                # We append the the colomn separator to the header size.
                header_size += column_separator

            if header_separator:
                # The header separator is given.

                # We append the right size of separator to the list of
                # separator data.
                header_separator_data.append(header_separator * size)

            # We increase the iterator.
            i += 1

        if header_separator:
            # The header separator is given.

            return [
                # We return the formatted header (like we will do with print('%s' % 'hello'))
                header_size % tuple(header_data),
                # We return the formatted header separator.
                header_size % tuple(header_separator_data),
            ]

        # The header separator is not given.

        # We return the formetted header.
        return [header_size % tuple(header_data)]

    def header(
        self, do_not_print=False
    ):  # pragma: no cover pylint: disable=too-many-branches
        """
        Management and creation of templates of header.
        Please consider as "header" the title of each columns.

        :param bool do_not_print:
            Tell us if we have to print the header or not.
        """

        if (
            not PyFunceble.CONFIGURATION.header_printed
            or self.template == "Percentage"
            or do_not_print
        ):
            # * The header has not been already printed.
            # or
            # * The template is the `Percentage template`.
            # or
            # * We are authorized to print something.

            if (
                self.template.lower() in PyFunceble.STATUS.list.generic
                or self.template == "Generic_File"
            ):
                # * The template is into the list of generic status.
                # or
                # * The template is equal to `Generic_File`.

                # The data to print is the Generic header.
                to_print = self.headers["Generic"]

                if (
                    self.template.lower() in PyFunceble.STATUS.list.generic
                    and PyFunceble.HTTP_CODE.active
                ):
                    # * The template is in the list of generic status.
                    # and
                    # * the http status code extraction is activated.

                    # We remove the Analyze Date colomn from the data to print.
                    to_print = PyFunceble.helpers.Dict(to_print).remove_key(
                        "Analyze Date"
                    )
            elif self.template.lower() in PyFunceble.STATUS.list.up:
                # The template is in the list of up status.

                # We informations to print is the up header.
                to_print = self.headers[PyFunceble.STATUS.official.up]
            elif self.template.lower() in PyFunceble.STATUS.list.valid:
                # The template is in the list of valid status.

                # We informations to print is the valid header.
                to_print = self.headers[PyFunceble.STATUS.official.valid]
            elif self.template.lower() in PyFunceble.STATUS.list.sane:
                # The template is in the list of sane status.

                # We informations to print is the sane header.
                to_print = self.headers[PyFunceble.STATUS.official.sane]
            elif self.template.lower() in PyFunceble.STATUS.list.malicious:
                # The template is in the list of malicious status.

                # We informations to print is the malicious header.
                to_print = self.headers[PyFunceble.STATUS.official.malicious]
            elif self.template.lower() in PyFunceble.STATUS.list.down:
                # The template is in the list of down status.

                # We informations to print is the down header.
                to_print = self.headers[PyFunceble.STATUS.official.down]
            elif self.template.lower() in PyFunceble.STATUS.list.invalid:
                # The template is in the list of invalid status.

                # We informations to print is the invalid header.
                to_print = self.headers[PyFunceble.STATUS.official.invalid]
            elif (
                self.template == "Less"
                or self.template == "Percentage"
                or self.template == "HTTP"
            ):  # pylint: disable=line-too-long
                # * The template is equal to `Less`.
                # or
                # * The template is equal to `Percentage`.
                # or
                # * The template is equal to `HTTP`.

                # We get the header with the help of the template name.
                to_print = self.headers[self.template]

                if self.template == "Less" and not PyFunceble.HTTP_CODE.active:
                    # * The template is equal to `Less`.
                    # and
                    # * The http status code extraction is deactivated.

                    # We append the source index to the header.
                    to_print["Source"] = 10
            elif self.template == "Simple":
                to_print = self.headers[self.template]

            if not PyFunceble.HTTP_CODE.active:
                # * The http status code extraction is deactivated.

                # We remove the HTTP Code index from the data to print.
                to_print = PyFunceble.helpers.Dict(to_print).remove_key("HTTP Code")

            # We update the currently used header.
            self.currently_used_header = to_print

            if not do_not_print:
                # We are not authorized to print anything.

                # We generate the before header.
                self.before_header()

                for formatted_template in self.header_constructor(to_print):
                    # We loop through the formatted template.

                    if not self.only_on_file:
                        # We do not have to print only on file.

                        # We print on screen the formatted header template.
                        print(formatted_template)

                    if not PyFunceble.CONFIGURATION.no_files and self.output:
                        # An output destination is given.

                        # We write the file with the formatted header template.
                        self.file_output_instance.write(formatted_template + "\n")

    def data_constructor(self, size):
        """
        Construct the table of data according to given size.

        :param list size: The maximal length of each string in the table.

        :return:
            A dict with all information about the data and how to which what
            maximal size to print it.
        :rtype: OrderedDict

        :raises:
            :code:`Exception`
                If the data and the size does not have the same length.
        """

        # We initiate a variable which will save what we are going to
        # return.
        result = OrderedDict()

        if len(self.data_to_print) == len(size):
            # The length of the data to print is equal to the length of the given size.

            for i in range(len(self.data_to_print)):
                # We loop until our iterator is less or equal to the length of the data
                # to print.

                # We initiate the result index and its size.
                result[self.data_to_print[i]] = size[i]
        else:
            # This should never happend. If it's happens then there is something
            # wrong from the inputed data.
            raise Exception(
                "Inputed: " + str(len(self.data_to_print)) + "; Size: " + str(len(size))
            )

        # We return the constructed result.
        return result

    @classmethod
    def _size_from_header(cls, header):
        """
        Get the size of each columns from the header.

        :param dict header:
            The header template we have to get the size from.

        :return: The maximal size of the each data to print.
        :rtype: list
        """

        # We initiate the result we are going to return.
        result = []

        for data in header:
            # We lopp through the header.

            # And we append the size to our result.
            result.append(header[data])

        # We return the result.
        return result

    def _colorify(self, data):
        """
        Retun colored string.

        :param str data: The string to colorify.

        :return: A colored string.
        :rtype: str
        """

        if self.template in ["Generic", "Less"]:
            # The template is in the list of template that need the coloration.

            if (
                self.data_to_print[1].lower() in PyFunceble.STATUS.list.up
                or self.data_to_print[1].lower() in PyFunceble.STATUS.list.valid
                or self.data_to_print[1].lower() in PyFunceble.STATUS.list.sane
            ):
                # The status is in the list of up status.

                # We print the data with a green background.
                data = Fore.BLACK + Back.GREEN + data
            elif (
                self.data_to_print[1].lower() in PyFunceble.STATUS.list.down
                or self.data_to_print[1].lower() in PyFunceble.STATUS.list.malicious
            ):
                # The status is in the list of down status.

                # We print the data with a red background.
                data = Fore.BLACK + Back.RED + data
            else:
                # The status is not in the list of up and down status.

                # We print the data with a cyan background.
                data = Fore.BLACK + Back.CYAN + data

        # We return the data.
        return data

    def _json_print(self):  # pragma: no cover
        """
        Management of the json template.
        """

        if self.output:
            # The given output is not empty.

            if self.file_output_instance.exists():
                # The given output already exist.

                # We get the content of the output.
                content = PyFunceble.helpers.Dict().from_json_file(
                    self.file_output_instance.path
                )

                if not content or isinstance(content, dict):
                    content = []

                if isinstance(content, list):
                    # The content is a list.

                    # We extend the content with our data to print.
                    content.extend(self.data_to_print)

                    # We format our list.
                    content = PyFunceble.helpers.List(content).custom_format(
                        PyFunceble.engine.Sort.standard
                    )

                    if PyFunceble.CONFIGURATION.hierarchical_sorting:
                        # The hierarchical sorting is activated.

                        # We format our content hierarchicaly
                        content = PyFunceble.helpers.List(content).custom_format(
                            PyFunceble.engine.Sort.hierarchical
                        )

                    # We finally save our content into the file.
                    PyFunceble.helpers.Dict(content).to_json_file(self.output)
                else:
                    # The content is not a list.

                    # We raise an exception.
                    raise Exception("Output not correctly formatted.")
            else:
                # The given output does not already exist.

                # We save our data to print into the output.
                #
                # Note: We do not have to take care if self.data_to_print is a list
                # formatted or not because this method should not be called if it is
                # not the case.
                PyFunceble.helpers.Dict(self.data_to_print).to_json_file(self.output)
        else:
            # The given output is empty.

            # We raise an exception.
            raise Exception("Empty output given.")

    def __get_print_size(self):  # pragma: no cover
        """
        Provides the size of the element to print.
        """

        # We initiate a variable which will list the list of
        # alone case.
        alone_cases = ["Percentage", "HTTP"]

        # we initiate a variable which will list the list of
        # template which does not need a header.
        without_header = ["FullHosts", "PlainDomain"]

        result = []

        if self.template not in alone_cases and self.template not in without_header:
            # * The template is not in the list of alone case.
            # and
            # * THe template is not in the list of template without header.

            # We get the template we should use.
            # Note: We basically only need the self.currently_used_header to be filled.
            self.header(True)

            # And we get the size from the header.
            return self._size_from_header(self.currently_used_header)

        if self.template in without_header:
            # The template is in the list of template which does not need a header.

            result = []
            for data in self.data_to_print:
                # We loop through the list of data to print.

                # And we construct the (spacement) size of the data to print.
                result.append(str(len(data)))
            return result

        # We get the size from the given template name.
        return self._size_from_header(self.headers[self.template])

    def data(self):  #  pragma: no cover  pylint: disable=inconsistent-return-statements
        """
        Management and input of data to the table.

        :raises:
            :code:`Exception`
                When self.data_to_print is not a list.
        """

        if isinstance(self.data_to_print, list):
            # The data to print is a list.

            # We initiate the data we are going to print.
            to_print = {}

            if self.template.lower() == "json":
                # The template is the json template.

                if not PyFunceble.CONFIGURATION.no_files and self.output:
                    # * We are allowed to generate file.
                    # and
                    # * The given output is not empty.

                    # We print the json file.
                    return self._json_print()

                # We return nothing.
                return None

            # We initiate the size we are going to print.
            to_print_size = self.__get_print_size()

            # We construct and format the data to print.
            to_print = self.data_constructor(to_print_size)

            # We print the before header section.
            self.before_header()

            for data in self.header_constructor(to_print, False):
                # We loop through the formatted data.

                if self.template.lower() in PyFunceble.STATUS.list.generic or self.template in [
                    "Less",
                    "Percentage",
                ]:
                    # * The template is in the list of generic status.
                    # or
                    # * The template is in a specific list.

                    if not self.only_on_file:
                        # We are authorized to print on screen.

                        # We colorify the data to print.
                        colorified_data = self._colorify(data)

                        # And we print the data.
                        print(colorified_data)

                elif self.template == "Simple":
                    if not self.only_on_file:
                        print(
                            PyFunceble.core.CLI.get_simple_coloration(
                                self.data_to_print[-1]
                            )
                            + data
                            + Style.RESET_ALL
                        )

                if not PyFunceble.CONFIGURATION.no_files and self.output:
                    # * We are authorized to print on any file.
                    # and
                    # * The output is given.

                    # We write our data into the printed file.
                    self.file_output_instance.write(data + "\n")
        else:
            # This should never happend. If it's happens then there's a big issue
            # around data_to_print.
            raise Exception("Please review Prints().data()")



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our sorting engines.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import PyFunceble


class Sort:  # pylint: disable=too-few-public-methods
    """
    Provides some sorting presets which we can
    parse to :py:class:`PyFunceble.helpers.list.List.custom_format`.
    """

    # We initiate a regex which will match everything which is not
    # a letter or a number.
    regex_replace = r"[^a-zA-Z0-9]"

    @classmethod
    def standard(cls, element):
        """
        Implements the standard and alphabetical sorting.

        :param str element: The element we are currently reading.

        :return: The formatted element.
        :rtype: str
        """

        # We remove all special characters and return the formatted string.
        return (
            PyFunceble.helpers.Regex(cls.regex_replace)
            .replace_match(element.strip(), "@funilrys")
            .replace("@funilrys", "")
        )

    @classmethod
    def hierarchical(cls, element):
        """
        The idea behind this method is to sort a list of domain hierarchicaly.

        :param str element: The element we are currently reading.

        :return: The formatted element.
        :rtype: str

        .. note::
            For a domain like :code:`aaa.bbb.ccc.tdl`.

            A normal sorting is done in the following order:
                1. :code:`aaa`
                2. :code:`bbb`
                3. :code:`ccc`
                4. :code:`tdl`

            This method allow the sorting to be done in the following order:
                1. :code:`tdl`
                2. :code:`ccc`
                3. :code:`bbb`
                4. :code:`aaa`

        """

        # We initiate a variable which will save the element to sort without
        # the extension.
        to_sort = ""

        # We initiate a variable which will save the full extension.
        full_extension = ""

        # We convert the parsed element to lower case.
        element = element.lower().strip()

        # We try to get the url base.
        url_base = PyFunceble.Check(element).is_url(return_base=True)

        if not isinstance(url_base, str):
            # The url base is not found.

            if "." in element:
                # There is point in the parsed element.

                # We get the position of the first letter of the extension.
                extension_index = element.rindex(".") + 1

                # We get the extension from the position of the first letter
                # of the extension.
                extension = element[extension_index:]

                if extension in PyFunceble.PSLOOOKUP:
                    # The extension is in the public suffix database.

                    for suffix in PyFunceble.PSLOOOKUP[extension]:
                        # We loop through the list of suffix of the extracted extension.

                        # We suffix the sufix with a point.
                        formatted_suffix = "." + suffix

                        if element.endswith(formatted_suffix):
                            # The elements ends with the suffix.

                            # We get the position of the first character of the suffix in
                            # the parsed element.
                            suffix_index = element.rindex(formatted_suffix)

                            # We update the to_sort variable with the element without the suffix.
                            to_sort = element[:suffix_index]

                            # We replace the full extension with the currently read suffix.
                            full_extension = suffix

                            # We break the loop, we got what we wanted.
                            break

                if not full_extension:
                    # The full extension is empty.

                    # We initiate it with the extension.
                    full_extension = element[extension_index:]

                    # We update the to_sort variable with the element without the extension.
                    to_sort = element[: extension_index - 1]

                # We append a point to the full extension because the point has to be
                # at the end and not at the begining of the extension.
                # To understand: Imagine a miror.
                full_extension += "."

                # We reverse the to_sort string.
                tros_ot = to_sort[::-1]

                if "." in tros_ot:
                    # There is a point in the reversed string.

                    # We prefix the full extension with the top level
                    # domain name.
                    full_extension = (
                        tros_ot[: tros_ot.index(".")][::-1] + "." + full_extension
                    )

                    # We remove the tor level domain from the rest of
                    # the reversed string.
                    tros_ot = tros_ot[tros_ot.index(".") + 1 :]

                    # * We reverse each level of the parsed element.
                    # and
                    # * We glue each level of the parsed element with each other.
                    #
                    # Note: after this, there is no point anymore.
                    reversion = full_extension + ".".join(
                        [x[::-1] for x in tros_ot.split(".")]
                    )

                    # We remove all special characters and return the formatted string.
                    return (
                        PyFunceble.helpers.Regex(cls.regex_replace)
                        .replace_match(reversion, "@funilrys")
                        .replace("@funilrys", "")
                    )

                # We remove all special characters and return the formatted string.
                return (
                    PyFunceble.helpers.Regex(cls.regex_replace)
                    .replace_match(to_sort + full_extension, "@funilrys")
                    .replace("@funilrys", "")
                )

            # There is no point in the parsed element.

            # We return the parsed element.
            return element

        # The url base is found.

        # We get the position of the element.
        protocol_position = element.rindex(url_base)

        # We extract the protocol from the element position.
        protocol = element[:protocol_position]

        # We return the output of this method but with the url base instead of the full url.
        return protocol + cls.hierarchical(url_base)



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the WHOIS database interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from datetime import datetime

import PyFunceble


class WhoisDB:
    """
    Provides the WHOIS database interface and logic.
    """

    database = {}

    database_file = None
    authorized = False

    def __init__(self, mysql_db=None, parent_process=False):
        # Get the authorization.
        self.authorized = self.authorization()
        self.database_file = ""

        if PyFunceble.CONFIGURATION.db_type == "json":
            # We set the location of the database file.
            self.database_file = "{0}{1}".format(
                PyFunceble.CONFIG_DIRECTORY, PyFunceble.OUTPUTS.default_files.whois_db
            )

        self.parent = parent_process
        self.mysql_db = mysql_db
        self.table_name = self.get_table_name()

        PyFunceble.LOGGER.debug(f"DB: {self.mysql_db}")
        PyFunceble.LOGGER.debug(f"Table Name: {self.table_name}")
        PyFunceble.LOGGER.debug(f"DB (File): {self.database_file}")

        # We load the configuration.
        self.load()

    def __contains__(self, index):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                if index in self.database:
                    PyFunceble.LOGGER.info(f"{index} is present into the database.")
                    return True

                PyFunceble.LOGGER.info(f"{index} is not present into the database.")
                return False

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = "SELECT COUNT(*) FROM {0} WHERE subject = %(subject)s".format(
                    self.table_name
                )

                with self.mysql_db.get_connection() as cursor:
                    cursor.execute(query, {"subject": index})

                    fetched = cursor.fetchone()

                    if fetched["COUNT(*)"] != 0:
                        PyFunceble.LOGGER.info(f"{index} is present into the database.")
                        return True

                    PyFunceble.LOGGER.info(f"{index} is not present into the database.")

        return False  # pragma: no cover

    def __getitem__(self, index):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                if index in self.database:
                    return self.database[index]

                return None

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                fetched = None

                if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                    query = "SELECT * FROM {0} WHERE subject = %(subject)s".format(
                        self.table_name
                    )

                    with self.mysql_db.get_connection() as cursor:
                        cursor.execute(query, {"subject": index})

                        fetched = cursor.fetchone()

                if fetched:
                    return {
                        "epoch": fetched["expiration_date_epoch"],
                        "expiration_date": fetched["expiration_date"],
                        "state": fetched["state"],
                    }

        return None  # pragma: no cover

    def __setitem_json(self, index, value):
        actual_value = self[index]

        if "record" in value:
            del value["record"]

        if isinstance(actual_value, dict):
            if isinstance(value, dict):
                self.database[index].update(value)
            else:  # pragma: no cover
                self.database[index] = value
        elif isinstance(actual_value, list):
            if isinstance(value, list):  # pragma: no cover
                self.database[index].extend(value)
            else:  # pragma: no cover
                self.database[index].append(value)
        else:
            self.database[index] = value

        PyFunceble.LOGGER.info(
            f"Inserted {repr(value)} into the subset of {repr(index)}"
        )

    def __setitem_mysql(self, index, value):
        query = (
            "INSERT INTO {0} "
            "(subject, expiration_date, expiration_date_epoch, state, record, digest) "
            "VALUES  (%(subject)s, %(expiration_date)s, %(epoch)s, %(state)s, %(record)s, %(digest)s)"  # pylint: disable=line-too-long
        ).format(self.table_name)

        digest = PyFunceble.helpers.Hash(algo="sha256").data(
            bytes(
                index + value["expiration_date"] + str(value["epoch"]) + value["state"],
                "utf-8",
            )
        )

        with self.mysql_db.get_connection() as cursor:
            playload = {
                "subject": index,
                "expiration_date": value["expiration_date"],
                "epoch": value["epoch"],
                "state": value["state"],
                "record": value["record"],
                "digest": digest,
            }
            try:
                cursor.execute(query, playload)
                PyFunceble.LOGGER.info(f"Inserted into the database: \n {playload}")
            except self.mysql_db.errors:
                pass

    def __setitem__(self, index, value):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                self.__setitem_json(index, value)
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                self.__setitem_mysql(index, value)

    @classmethod
    def authorization(cls):
        """
        Provides the operation authorization.
        """

        return (
            not PyFunceble.CONFIGURATION.no_whois
            and PyFunceble.CONFIGURATION.whois_database
        )

    @classmethod
    def merge(cls, old):  # pragma: no cover
        """
        Merges the older version of the database into the new version.

        :param dict old: The old version of the database.


        :return: The database in the new format.
        :rtype: dict
        """

        # We initiate a local place to save our results.
        result = {}

        if PyFunceble.CONFIGURATION.db_type == "json":
            for index, data in old.items():
                # We loop through all indexes and data of the database.

                if isinstance(data, dict) and "epoch" in data:
                    # The epoch index is present into the currently
                    # read dataset.

                    # We create the copy of the dataset for our result.
                    result[index] = data

                    continue

                if isinstance(data, dict):
                    # The read data is a dict.

                    # We save the content of of the currently read dataset
                    # into the upstream index.
                    result.update(data)

        # We return the result.
        return result

    def get_table_name(self):
        """
        Returns the name of the table to use.
        """

        if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
            return self.mysql_db.tables["whois"]
        return "whois"

    def load(self):
        """
        Loads the database file into the database.
        """

        if (
            self.authorized
            and PyFunceble.helpers.File(self.database_file).exists()
            and PyFunceble.CONFIGURATION.db_type == "json"
        ):
            # * We are authorized to operate.
            # and
            # * The database file exists.

            # We merge our current database into already initiated one.
            self.database.update(
                self.merge(PyFunceble.helpers.Dict().from_json_file(self.database_file))
            )

            PyFunceble.LOGGER.info(
                "Database content loaded in memory. (DATASET WONT BE LOGGED)"
            )

            # As changes can happen because of the merging, we directly saved
            # the loaded data.
            self.save()

    def save(self):
        """
        Saves the database into the database file.
        """

        if (
            self.authorized
            and self.parent
            and PyFunceble.CONFIGURATION.db_type == "json"
        ):
            # We are authorized to operate.

            # We save the current state of the datbase.
            PyFunceble.helpers.Dict(self.database).to_json_file(self.database_file)

            PyFunceble.LOGGER.info(f"Saved database into {repr(self.database_file)}.")

    def is_time_older(self, subject):
        """
        Checks if the expiration time of the given subject is
        older.

        :param str subject: The subject we are working with.

        .. note::
            Of course, we imply that the subject is in the database.
        """

        data = self[subject]

        return (
            self.authorized
            and data
            and "epoch" in data
            and datetime.fromtimestamp(float(data["epoch"])) < datetime.now()
        )

    def get_expiration_date(self, subject):
        """
        Gets the expiration date of the given subject.

        :param str subject: The subject we are working with.

        :return: The expiration date from the database.
        :rtype: str|None
        """

        if self.authorized and self[subject] and not self.is_time_older(subject):
            # * We are authorized to work.
            # and
            # * The element we are testing is in the database.
            # and
            # * The expiration date is in the future.

            try:
                # We return the expiration date.
                return self[subject]["expiration_date"]
            except KeyError:  # pragma: no cover
                pass

        # We return None, there is no data to work with.
        return None

    def add(self, subject, expiration_date, record=None):
        """
        Adds the given subject and expiration date to the database.

        :param str subject: The subject we are working with.
        :param str expiration_date: The extracted expiration date.
        :param str record: The WHOIS record.
        """

        if self.authorized and expiration_date:
            # * We are authorized to operate.
            # and
            # * The expiration date is not empty nor None.

            # We initiate what we are going to save into the database
            data = {
                "epoch": datetime.strptime(expiration_date, "%d-%b-%Y").timestamp(),
                "expiration_date": expiration_date,
            }

            if record:  # pragma: no cover
                data["record"] = str(record)
            else:
                data["record"] = str(None)

            if data["epoch"] < datetime.now().timestamp():
                # We compare the epoch with the current time.

                # We set the state.
                data["state"] = "past"
            else:
                # We set the state.
                data["state"] = "future"

            # We save everything into the database.
            self[subject] = data

            # We save everything.
            self.save()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.converters.file

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
# pylint: enable=line-too-long
from unittest import TestCase
from unittest import main as launch_tests

import PyFunceble
from PyFunceble.converter.file import File


class TestFileLineConverter(TestCase):
    """
    Tests of PyFunceble.converter.file
    """

    def setUp(self):
        """
        Setups everything that is needed for the tests.
        """

        PyFunceble.load_config(generate_directory_structure=False)

        self.domains = [
            "google.com",
            "twitter.com",
            "github.com",
            "facebook.com",
            "hello.world",
            "world.hello",
        ]

    def tests_simple_line(self):
        """
        Tests the case that we encouter a simple line.
        """

        for domain in self.domains:
            expected = domain
            actual = File(domain).get_converted()

            self.assertEqual(expected, actual)

    def tests_comment(self):
        """
        Tests the case that we encouter a commented line.
        """

        for domain in self.domains:
            expected = None

            data = f"# {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_ends_with_comment(self):
        """
        Tests the case that a line has a comment at the end of its line.
        """

        for domain in self.domains:
            expected = domain

            data = f"{domain} # hello world"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_prefix(self):
        """
        Tests the case that a line has a decorator.

        For example:
        ::

            127.0.0.1 google.com
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0 {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1 {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_prefix_and_multiple_subjects(self):
        """
        Tests the case that a line has a decorator and multiple subjects.

        For example:
        ::

            127.0.0.1 google.com example.com example.net
        """

        for domain in self.domains:
            expected = [domain, "example.com", "example.net"]

            data = f"0.0.0.0 {domain} example.com\t\texample.net"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = [domain, "example.com", "example.net"]

            data = f"127.0.0.1 {domain} example.com\t\texample.net"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_multiple_spaces(self):
        """
        Tests the case that we have multiple space as sparator between
        our domain end its prefix.
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0                {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1                {domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_tabs(self):
        """
        Tests the case that we have a single tab as sparator between
        our domain end its prefix.
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_multiple_tabs(self):
        """
        Tests the case that we have multiple tabs as sparator between
        our domain end its prefix.
        """

        for domain in self.domains:
            expected = domain

            data = f"0.0.0.0\t\t\t\t\t\t\t\t\t\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

        for domain in self.domains:
            expected = domain

            data = f"127.0.0.1\t\t\t\t\t\t\t\t\t\t\t{domain}"
            actual = File(data).get_converted()

            self.assertEqual(expected, actual)

    def tests_with_list_of_entries(self):
        """
        Tests the case that we give multiple entries
        """

        given = [f"0.0.0.0\t\t\t\t\t\t\t\t\t\t{domain}" for domain in self.domains]
        expected = self.domains
        actual = File(given).get_converted()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.helpers.file

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
# pylint: enable=line-too-long

from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.helpers import File


class TestFile(TestCase):
    """
    Tests of PyFunceble.helpers.file.
    """

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        self.file = "this_file_is_a_ghost"

        self.file_instance = File(self.file)
        self.file_instance_2 = File(self.file + "_2")

        self.file_instance.delete()
        self.file_instance_2.delete()

    def tearDown(self):
        """
        Setups everything needed after the tests.
        """

        self.file_instance.delete()
        self.file_instance_2.delete()

    def test_exists(self):
        """
        Tests method which let us check the existance
        of a file.
        """

        expected = False
        actual = self.file_instance.exists()

        self.assertEqual(expected, actual)

        self.file_instance.write(self.file_instance.path)

        expected = True
        actual = self.file_instance.exists()

        self.assertEqual(expected, actual)

    def test_get_size(self):
        """
        Tests the method which let us get the size of a file.
        """

        self.file_instance.write("Hello, World!", overwrite=True)

        expected = 13
        actual = self.file_instance.get_size()

        self.assertEqual(expected, actual)

        self.file_instance.write(" ", overwrite=True)

        expected = 1
        actual = self.file_instance.get_size()

        self.assertEqual(expected, actual)

    def test_is_emtpy(self):
        """
        Tests the method which let us know if a file is empty.
        """

        self.file_instance.write("Hello!!", overwrite=True)

        expected = False
        actual = self.file_instance.is_empty()

        self.assertEqual(expected, actual)

        self.file_instance.write("", overwrite=True)

        expected = True
        actual = self.file_instance.is_empty()

        self.assertEqual(expected, actual)

    def test_delete_read_and_write(self):
        """
        Tests the method which let us delete and write into a file.
        """

        self.file_instance.write("Hello, World!")

        expected = "Hello, World!"
        actual = self.file_instance.read()

        self.assertEqual(expected, actual)

        self.file_instance.write("Hello, World!")

        expected += "Hello, World!"
        actual = self.file_instance.read()

        self.assertEqual(expected, actual)

        self.file_instance.write("Crap!", overwrite=True)

        expected = "Crap!"
        actual = self.file_instance.read()

        self.assertEqual(expected, actual)

    def test_copy(self):
        """
        Tests the method which let us copy a file to another location.
        """

        self.file_instance.write("Hello, World!")

        expected = "Hello, World!"
        actual = self.file_instance.read()

        self.assertEqual(expected, actual)

        self.file_instance.copy(self.file_instance_2.path)

        expected = True

        self.assertEqual(expected, self.file_instance_2.exists())
        self.assertEqual(expected, self.file_instance.exists())

        expected = "Hello, World!"
        actual = self.file_instance_2.read()

        self.assertEqual(expected, actual)

    def test_move(self):
        """
        Tests the method which let us move a file to another location.
        """

        self.file_instance.write("Hello, World!")

        expected = "Hello, World!"
        actual = self.file_instance.read()

        self.assertEqual(expected, actual)

        self.file_instance.move(self.file_instance_2.path)

        expected = True
        actual = self.file_instance_2.exists()

        self.assertEqual(expected, actual)

        expected = False
        actual = self.file_instance.exists()

        self.assertEqual(expected, actual)

        expected = "Hello, World!"
        actual = self.file_instance_2.read()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long,invalid-name,import-error
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.cli.execution_time

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
# pylint: enable=line-too-long

import sys
from collections import OrderedDict
from datetime import datetime, timedelta
from unittest import main as launch_tests
from unittest.mock import patch

from colorama import Fore, Style
from colorama import init as initiate_colorama

import PyFunceble
from PyFunceble.cli.execution_time import ExecutionTime
from stdout_base import StdoutBase


class TestExecutionTime(StdoutBase):
    """
    Tests of PyFunceble.cli.execution time
    """

    # pylint: disable=unsupported-assignment-operation, unsubscriptable-object

    def setUp(self):
        """
        Setups everything needed for the tests
        """

        PyFunceble.load_config(generate_directory_structure=False)

        initiate_colorama(True)

        StdoutBase.setUp(self)

    def tearDown(self):
        """
        Setups everything needed for after the tests.
        """

        StdoutBase.tearDown(self)

    def test_authorization(self):
        """
        Tests the authorization method.
        """

        PyFunceble.CONFIGURATION.show_execution_time = False
        expected = False

        actual = ExecutionTime.authorization()

        self.assertEqual(expected, actual)

        PyFunceble.CONFIGURATION.show_execution_time = False
        PyFunceble.CONFIGURATION.ci = True
        expected = True

        actual = ExecutionTime.authorization()

        self.assertEqual(expected, actual)

        PyFunceble.CONFIGURATION.show_execution_time = False
        PyFunceble.CONFIGURATION.ci = False
        expected = False

        actual = ExecutionTime.authorization()

        self.assertEqual(expected, actual)

        PyFunceble.CONFIGURATION.show_execution_time = True
        PyFunceble.CONFIGURATION.ci = False
        expected = True

        actual = ExecutionTime.authorization()

        self.assertEqual(expected, actual)

    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_stoping_time", lambda x: x)
    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_starting_time", lambda x: x)
    def test_calculate(self):
        """
        Tests the calculation method.
        """

        PyFunceble.CONFIGURATION.show_execution_time = True

        start_time = datetime.now()
        PyFunceble.INTERN["start"] = start_time.timestamp()
        PyFunceble.INTERN["end"] = (start_time + timedelta(seconds=15)).timestamp()

        expected = OrderedDict(
            zip(["days", "hours", "minutes", "seconds"], ["00", "00", "00", "15.0"])
        )

        actual = ExecutionTime("stop").calculate()

        self.assertEqual(expected, actual)

        actual = ExecutionTime("stop").calculate(
            start=start_time.timestamp(),
            end=(start_time + timedelta(seconds=15)).timestamp(),
        )

        self.assertEqual(expected, actual)

    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_stoping_time", lambda x: x)
    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_starting_time", lambda x: x)
    def test_calculate_consequent(self):
        """
        Tests the calculation method with more consequent data.
        """

        start_time = datetime.now()
        PyFunceble.INTERN["start"] = start_time.timestamp()
        PyFunceble.INTERN["end"] = (
            start_time + timedelta(days=1, hours=50)
        ).timestamp()

        expected = OrderedDict(
            zip(["days", "hours", "minutes", "seconds"], ["03", "02", "00", "0.0"])
        )

        actual = ExecutionTime("stop").calculate()

        self.assertEqual(expected, actual)

        actual = ExecutionTime("stop").calculate(
            start=start_time.timestamp(),
            end=(start_time + timedelta(days=1, hours=50)).timestamp(),
        )

        self.assertEqual(expected, actual)

    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_stoping_time", lambda x: x)
    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_starting_time", lambda x: x)
    def test_save(self):
        """
        Test the saving method.
        """

        expected_file_location = (
            PyFunceble.OUTPUT_DIRECTORY
            + PyFunceble.OUTPUTS.parent_directory
            + PyFunceble.OUTPUTS.logs.directories.parent
            + PyFunceble.OUTPUTS.logs.filenames.execution_time
        )

        PyFunceble.helpers.File(expected_file_location).delete()

        PyFunceble.CONFIGURATION.show_execution_time = True
        PyFunceble.CONFIGURATION.logs = True
        PyFunceble.CONFIGURATION.show_percentage = False
        PyFunceble.INTERN["file_to_test"] = "this_is_a_ghost"

        start_time = datetime.now()
        PyFunceble.INTERN["start"] = start_time.timestamp()
        PyFunceble.INTERN["end"] = (start_time + timedelta(seconds=15)).timestamp()

        expected = {
            "current_total": "00:00:00:15.0",
            "data": [
                [PyFunceble.INTERN["start"], PyFunceble.INTERN["end"]],
                [PyFunceble.INTERN["start"], PyFunceble.INTERN["end"]],
            ],
            "final_total": "00:00:00:15.0",
        }

        ExecutionTime("start")
        ExecutionTime("stop")

        expected_stdout = (
            f"{Fore.MAGENTA}{Style.BRIGHT }\n"
            f"Execution time: {expected['final_total']}\n"
        )
        self.assertEqual(expected_stdout, sys.stdout.getvalue())

        ExecutionTime("start")
        ExecutionTime("stop", last=True)

        expected_stdout += expected_stdout
        expected_stdout += (
            f"{Fore.MAGENTA}{Style.BRIGHT }"
            f"Global execution time: {expected['final_total']}\n"
        )
        self.assertEqual(
            expected_stdout,
            sys.stdout.getvalue(),
            f"{repr(sys.stdout.getvalue())}\n{repr(expected_stdout)}",
        )

        actual = PyFunceble.helpers.Dict().from_json_file(expected_file_location)
        self.assertEqual(expected, actual)

        del expected["final_total"]

        ExecutionTime("start")
        ExecutionTime("stop", last=True)

        expected["data"].extend(
            [[PyFunceble.INTERN["start"], PyFunceble.INTERN["end"]]]
        )
        expected["final_total"] = "00:00:00:15.0"

        actual = PyFunceble.helpers.Dict().from_json_file(expected_file_location)
        self.assertEqual(expected, actual)

    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_stoping_time", lambda x: x)
    @patch("PyFunceble.cli.execution_time.ExecutionTime.set_starting_time", lambda x: x)
    @patch("PyFunceble.cli.execution_time.ExecutionTime.calculate")
    def test_format_execution_time(self, calculate):
        """
        Tests of the date formatter method.
        """

        calculate.return_value = OrderedDict(
            zip(["days", "hours", "minutes", "seconds"], ["00", "00", "00", "15.0"])
        )
        expected = "00:00:00:15.0"
        actual = ExecutionTime("stop").format_execution_time()

        self.assertEqual(expected, actual)

        calculate.return_value = OrderedDict(
            zip(["days", "hours", "minutes", "seconds"], ["03", "02", "00", "0.0"])
        )

        expected = "03:02:00:0.0"
        actual = ExecutionTime("stop").format_execution_time()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base of all our downloader.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from datetime import datetime, timedelta

import PyFunceble


class DownloaderBase:
    """
    Provides the downloader base.
    """

    # pylint: disable=no-member

    DOWNTIME_INDEX = None
    """
    Used to set/track the download time of the current file.

    :rtype: str
    """

    REDOWNLOAD_AFTER = 1
    """
    Used to set the redownload frequency in day.

    :rtype: int
    """

    def __init__(self):
        self.downtimes_file = PyFunceble.helpers.File(
            f"{PyFunceble.CONFIG_DIRECTORY}{PyFunceble.abstracts.Infrastructure.DOWN_FILENAME}"
        )

        self.all_downtimes = self.get_all_downtimes()

        if not hasattr(self, "destination"):
            raise PyFunceble.exceptions.NoDownloadDestinationGiven()

        if not hasattr(self, "download_link"):
            raise PyFunceble.exceptions.NoDownloadLinkGiven()

    def get_all_downtimes(self):
        """
        Provides all download times.

        :rtype: dict
        """

        return PyFunceble.helpers.Dict().from_json_file(self.downtimes_file.path)

    def get_downtime(self):
        """
        Provides the download timestamp of the current :code:`DOWN_TIME_INDEX`.

        :rtype: dict, None
        """

        if self.is_downtime_set():
            return self.all_downtimes[self.DOWNTIME_INDEX]

        return None

    def update_downtime(self):
        """
        Updates the current download time.
        """

        current_datetime = datetime.now()

        self.all_downtimes[self.DOWNTIME_INDEX] = {
            "iso": current_datetime.isoformat(),
            "timestamp": current_datetime.timestamp(),
        }

    def save_downtimes(self):
        """
        Saves all downtimes.
        """

        PyFunceble.helpers.Dict(self.all_downtimes).to_json_file(
            self.downtimes_file.path
        )

    def is_downtime_set(self):
        """
        Checks if the downtime is set for the current :code:`DOWNTIME_INDEX`.

        :rtype: bool
        """

        return (
            self.DOWNTIME_INDEX in self.all_downtimes
            and self.all_downtimes[self.DOWNTIME_INDEX]
            and all(
                [
                    x in self.all_downtimes[self.DOWNTIME_INDEX]
                    and self.all_downtimes[self.DOWNTIME_INDEX][x]
                    for x in ["iso", "timestamp"]
                ]
            )
        )

    def is_last_download_expired(self):
        """
        Checks if the last downloaded file is expired.
        """

        if (
            not PyFunceble.helpers.File(self.destination).exists()
            or not self.is_downtime_set()
        ):
            return True

        last_download = datetime.fromtimestamp(self.get_downtime()["timestamp"])

        if (
            self.REDOWNLOAD_AFTER <= 0
            and (datetime.now() - last_download).seconds < 3600
        ):
            return False

        if last_download + timedelta(days=self.REDOWNLOAD_AFTER) <= datetime.now():
            return True

        return False

    def process(self):
        """
        Process the download and returns the downloaded text.

        :rtype: str, None
        """

        if self.is_last_download_expired() and PyFunceble.helpers.Download(
            self.download_link
        ).text(destination=self.destination):
            self.update_downtime()
            self.save_downtimes()
